@@ -5,6 +5,8 @@ from __future__ import absolute_import
 import os
 import time
 
+from keras.optimizers import Adadelta, Adam, Adamax, Adagrad, RMSprop, SGD
+
 from aetros.backend import JobImage
 
 try:
@@ -21,17 +23,30 @@ import keras.layers.convolutional
 
 import numpy as np
 
-from aetros.utils.image import get_layer_vis_square
+from aetros.utils.image import get_layer_vis_square, get_layer_vis_square_raw, get_image_tales
 from .keras_model_utils import ensure_dir, get_total_params
 import six
 
+# compatibility with keras 1.x
+def image_data_format():
+    if hasattr(K, 'image_data_format'):
+        return K.image_data_format()
+
+    if K.image_dim_ordering() == 'th':
+        return 'channel_first'
+    else:
+        return 'channel_last'
+
+
 class KerasLogger(Callback):
-    def __init__(self, trainer, job_backend, general_logger):
+    def __init__(self, trainer, job_backend, general_logger, force_insights=False):
         self.params = {}
         super(KerasLogger, self).__init__()
         self.validation_per_batch = []
+        self.insight_layer = []
         self.ins = None
         self.insights_sample_path = None
+        self.force_insights = force_insights
 
         self.trainer = trainer
         self.current = {}
@@ -55,7 +70,11 @@ class KerasLogger(Callback):
         self.data_gathered = False
 
         self.accuracy_channel = None
+        self.all_losses = None
         self.loss_channel = None
+        self.learning_rate_channel = None
+
+        self.learning_rate_start = 0
 
         self.insights_x = None
         self.best_total_accuracy = 0
@@ -63,20 +82,28 @@ class KerasLogger(Callback):
     def on_train_begin(self, logs={}):
         self.start_time = time.time()
         self.last_batch_time = time.time()
-        self.trainer.set_status('TRAINING')
-        self.trainer.set_job_info('total_params', get_total_params(self.model))
+        self.job_backend.set_status('TRAINING')
+        self.job_backend.set_info('parameters', get_total_params(self.model))
+        self.job_backend.set_info('backend', K.backend())
+
+        if self.model.optimizer and hasattr(self.model.optimizer, 'get_config'):
+            config = self.model.optimizer.get_config()
+            self.job_backend.set_info('optimizer', type(self.model.optimizer).__name__)
+            for i, v in config.items():
+                self.job_backend.set_info('optimizer.'+str(i), v)
 
         self.current['epoch'] = 0
         self.current['started'] = self.start_time
-        self.trainer.set_job_info('current', self.current)
-        nb_sample = self.params['nb_sample']  # training samples total
-        nb_epoch = self.params['nb_epoch']  # training epoches total
+        self.job_backend.set_system_info('current', self.current)
 
-        self.current['nb_sample'] = nb_sample
-        self.current['nb_epoch'] = nb_epoch
+        #compatibility with keras 1.x
+        if 'epochs' not in self.params and 'nb_epoch' in self.params:
+            self.params['epochs'] = self.params['nb_epoch']
+        if 'samples' not in self.params and 'nb_sample' in self.params:
+            self.params['samples'] = self.params['nb_sample']
 
         xaxis = {
-            'range': [1, nb_epoch],
+            'range': [1, self.params['epochs']],
             'title': u'Epoch ⇢'
         }
         yaxis = {
@@ -85,28 +112,64 @@ class KerasLogger(Callback):
             'dtick': 10
         }
 
-        self.accuracy_channel = self.job_backend.create_channel('accuracy', main_graph=True, xaxis=xaxis, yaxis=yaxis)
+        traces = ['validation', 'training']
+        if len(self.model.output_layers) > 1:
+            traces = []
+            for output in self.model.output_layers:
+                traces.append(output.name+'_validation')
+                traces.append(output.name+'_training')
+
+        self.accuracy_channel = self.job_backend.create_channel('accuracy', main=True, traces=traces, kpi=True, max_optimization=True, xaxis=xaxis, yaxis=yaxis)
         self.loss_channel = self.job_backend.create_loss_channel('loss', xaxis=xaxis)
+        self.learning_rate_channel = self.job_backend.create_channel('learning rate', traces=['start', 'end'], xaxis=xaxis)
+        self.job_backend.progress(0, self.params['epochs'])
+        if len(self.model.output_layers) > 1:
+            loss_traces = []
+            for output in self.model.output_layers:
+                loss_traces.append(output.name+'_validation')
+                loss_traces.append(output.name+'_training')
+
+            self.all_losses = self.job_backend.create_channel('All loss', main=True, xaxis=xaxis, traces=loss_traces)
+
+        images = self.build_insight_images()
+        self.job_backend.job_add_insight(0, images, None)
 
     def on_batch_begin(self, batch, logs={}):
 
         if not self.data_gathered:
+            validation_samples = self.trainer.nb_val_samples
+            if 'samples' in self.params:
+                training_samples = self.params['samples']
+                if hasattr(self.model, 'validation_data') and self.model.validation_data:
+                    validation_samples = len(self.model.validation_data[0])
+            else:
+                training_samples = self.params['steps'] * logs['size']
+                if self.trainer.is_generator(self.trainer.data_validation):
+                    validation_samples = self.trainer.nb_val_steps * logs['size']
+                elif self.trainer.data_validation:
+                    validation_samples = len(self.trainer.data_validation[0])
+
             # we need to do it in on_batch_begin due to the fact that
             # self.model.validation_data is not availabe in on_train_begin
             self.data_gathered = True
             dataset_infos = {}
             dataset_info = {
-                'Training': self.params['nb_sample'],
-                'Validation': len(self.model.validation_data[0]) if self.model.validation_data else self.trainer.nb_val_samples,
+                'Training': training_samples,
+                'Validation': validation_samples,
             }
             dataset_infos['input1'] = dataset_info
-            self.trainer.set_job_info('datasets', dataset_infos)
+            self.job_backend.set_system_info('datasets', dataset_infos)
 
-        batch_size = logs['size']
-        nb_batches = math.ceil(self.current['nb_sample'] / batch_size)  # normal nb batches
+        if 'nb_batches' not in self.current:
+            batch_size = logs['size']
+            if 'samples' in self.params:
+                nb_batches = math.ceil(self.params['samples'] / batch_size)  # normal nb batches
+            else:
+                nb_batches = self.params['steps']
 
-        self.current['nb_batches'] = nb_batches
-        self.current['batch_size'] = batch_size
+            self.current['nb_batches'] = nb_batches
+            self.current['batch_size'] = batch_size
+            self.job_backend.set_info('Batch size', batch_size)
 
     def on_batch_end(self, batch, logs={}):
         self.filter_invalid_json_values(logs)
@@ -114,42 +177,15 @@ class KerasLogger(Callback):
 
         self.validation_per_batch.append(loss)
 
-        current_batch = logs['batch']
-        # how many training items in this batch, differs for the last run
-        current_batch_size = logs['size']
-
-        self.made_batches += 1
-
-        time_diff = time.time() - self.last_batch_time
-
-        if time_diff > 2 or batch == self.current['nb_batches']:  # only each second second or last batch
-            self.batches_per_second = self.made_batches / time_diff
-            self.made_batches = 0
-            self.last_batch_time = time.time()
-
-            nb_sample = self.params['nb_sample']  # training samples total
-            batch_size = self.current['batch_size']  # normal batch size
-            nb_batches = nb_sample / batch_size  # normal nb batches
-
-            self.current['batchesPerSecond'] = self.batches_per_second
-            self.current['itemsPerSecond'] = self.batches_per_second * current_batch_size
-
-            epochs_per_second = self.batches_per_second / nb_batches  # all batches
-            self.current['epochsPerSecond'] = epochs_per_second
-
-            self.current['currentBatch'] = current_batch
-            self.current['currentBatchSize'] = current_batch_size
-            elapsed = time.time() - self.start_time
-            self.current['elapsed'] = elapsed
-
-            self.trainer.set_job_info('itemsPerSecond', self.batches_per_second * current_batch_size)
-            self.trainer.set_job_info('currentBatch', current_batch)
-            self.trainer.set_job_info('currentBatchSize', batch_size)
-            self.trainer.set_job_info('nb_batches', nb_batches)
+        self.job_backend.batch(batch, self.current['nb_batches'], logs['size'])
 
     def write(self, line):
         self.general_logger.write(line)
 
+    def on_epoch_begin(self, epoch, logs={}):
+        self.learning_rate_start = self.get_learning_rate()
+
+
     def on_epoch_end(self, epoch, logs={}):
         log = logs.copy()
 
@@ -160,26 +196,61 @@ class KerasLogger(Callback):
         if 'loss' not in log and len(self.validation_per_batch) > 0:
             log['loss'] = sum(self.validation_per_batch) / float(len(self.validation_per_batch))
 
-        total_accuracy = log.get('val_acc', 0)
+        accuracy_log_name = 'acc'
+        val_accuracy_log_name = 'val_acc'
 
-        if total_accuracy > self.best_total_accuracy:
-            self.best_total_accuracy = total_accuracy
+        if len(self.model.output_layers) > 1:
+            name = self.model.output_layers[0].name
+            accuracy_log_name = name+'_acc'
+            val_accuracy_log_name = 'val_' + name + '_acc'
+
+        total_accuracy_validation = log.get(val_accuracy_log_name, 0)
+        total_accuracy_training = log.get(accuracy_log_name, 0)
+
+        if total_accuracy_validation > self.best_total_accuracy:
+            self.best_total_accuracy = total_accuracy_validation
             self.best_epoch = log['epoch']
-            self.model.save_weights(self.filepath_best, overwrite=True)
+            try:
+                self.model.save_weights(self.filepath_best, overwrite=True)
+            except:
+                # sometimes hangs with: IOError: Unable to create file (Unable to open file: name = ...
+                # without any obvious reason.
+                pass
 
-        self.model.save_weights(self.filepath_latest, overwrite=True)
+        try:
+            self.model.save_weights(self.filepath_latest, overwrite=True)
+        except:
+            # sometimes hangs with: IOError: Unable to create file (Unable to open file: name = ...
+            # without any obvious reason.
+            pass
 
         self.loss_channel.send(log['epoch'], log.get('loss', 0), log.get('val_loss', 0))
-        self.accuracy_channel.send(log['epoch'], total_accuracy*100)
 
-        self.job_backend.progress(log['epoch'], self.params['nb_epoch'])
+        accuracy = [total_accuracy_validation*100, total_accuracy_training*100]
+        if len(self.model.output_layers) > 1:
+            accuracy = []
+            losses = []
+            for layer in self.model.output_layers:
+                accuracy.append(log.get('val_' + layer.name + '_acc', 0)*100)
+                accuracy.append(log.get(layer.name + '_acc', 0)*100)
+
+                losses.append(log.get('val_' + layer.name + '_loss', 0))
+                losses.append(log.get(layer.name + '_loss', 0))
+
+            self.all_losses.send(log['epoch'], losses)
+
+        self.accuracy_channel.send(log['epoch'], accuracy)
+
+        self.job_backend.progress(log['epoch'], self.params['epochs'])
+        self.send_optimizer_info(log['epoch'])
 
         if self.log_epoch:
+            #todo, multiple outputs
             line = "Epoch %d: loss=%f, acc=%f, val_loss=%f, val_acc=%f\n" % (
-                log['epoch'], log['loss'], log.get('acc', 0), log.get('val_loss', 0), total_accuracy, )
+                log['epoch'], log['loss'], log.get('acc', 0), log.get('val_loss', 0), total_accuracy_validation, )
             self.general_logger.write(line)
 
-        if self.job_model.job['insights']:
+        if self.force_insights or self.job_model.job['insights']:
             # Todo, support multiple inputs
             first_input_layer = self.model.input_layers[0]
 
@@ -191,6 +262,22 @@ class KerasLogger(Callback):
 
                 self.job_backend.job_add_insight(log['epoch'], images, confusion_matrix)
 
+    def send_optimizer_info(self, epoch):
+        self.learning_rate_channel.send(epoch, [self.learning_rate_start, self.get_learning_rate()])
+
+    def get_learning_rate(self):
+
+        if hasattr(self.model, 'optimizer'):
+            config = self.model.optimizer.get_config()
+
+            if isinstance(self.model.optimizer, Adadelta) or isinstance(self.model.optimizer, Adam) \
+                    or isinstance(self.model.optimizer, Adamax) or isinstance(self.model.optimizer, Adagrad)\
+                    or isinstance(self.model.optimizer, RMSprop) or isinstance(self.model.optimizer, SGD):
+                return config['lr'] * (1. / (1. + config['decay'] * float(K.get_value(self.model.optimizer.iterations))))
+
+            elif 'lr' in config:
+                return config['lr']
+
     def is_image_shape(self, x):
         if len(x.shape) != 3 and len(x.shape) != 2:
             return False
@@ -206,7 +293,7 @@ class KerasLogger(Callback):
             return (x.shape[2] == 1 or x.shape[2] == 3)
 
     def get_first_input_sample(self):
-        if self.model.validation_data:
+        if hasattr(self.model, 'validation_data'):
             input_data_x = []
             for i, input in enumerate(self.model.inputs):
                 input_data_x.append([self.model.validation_data[i][0]])
@@ -241,14 +328,12 @@ class KerasLogger(Callback):
 
         images = []
 
-        try:
-            for i, layer in enumerate(self.model.input_layers):
-                x = np.squeeze(self.insights_x[i])
-                if self.is_image_shape(x):
-                    image = self.make_image(x)
+        for i, layer in enumerate(self.model.input_layers):
+            x = np.squeeze(self.insights_x[i])
+            if self.is_image_shape(x):
+                image = self.make_image(x)
+                if image:
                     images.append(JobImage(layer.name, image))
-        except:
-            pass
 
         uses_learning_phase = self.model.uses_learning_phase
         inputs = self.model.inputs[:]
@@ -258,54 +343,77 @@ class KerasLogger(Callback):
             inputs += [K.learning_phase()]
             input_data_x_sample += [0.]  # disable learning_phase
 
-        for layer in self.model.layers:
+        layers = self.model.layers + self.insight_layer
+
+        for layer in layers:
             if isinstance(layer, keras.layers.convolutional.Convolution2D) or isinstance(layer, keras.layers.convolutional.MaxPooling2D):
 
                 fn = K.function(inputs, self.get_layout_output_tensors(layer))
                 Y = fn(input_data_x_sample)[0]
 
                 data = np.squeeze(Y)
-                image = PIL.Image.fromarray(get_layer_vis_square(data))
+                if image_data_format() == 'channels_last':
+                    data = np.transpose(data, (2, 0, 1))
 
+                image = PIL.Image.fromarray(get_image_tales(data))
                 images.append(JobImage(layer.name, image))
 
-                if hasattr(layer, 'W') and layer.W:
+                if layer.get_weights():
                     data = layer.get_weights()[0]
-                    image = PIL.Image.fromarray(get_layer_vis_square(data))
+
+                    data = np.transpose(data, (2, 3, 0, 1))
+
+                    data = data.reshape((data.shape[0] * data.shape[1], data.shape[2], data.shape[3]))
+
+                    image = PIL.Image.fromarray(get_image_tales(data))
                     images.append(JobImage(layer.name + '_weights', image, layer.name + ' weights'))
 
-            if isinstance(layer, keras.layers.Dense):
+            elif isinstance(layer, keras.layers.ZeroPadding2D) or isinstance(layer, keras.layers.ZeroPadding1D) or isinstance(layer, keras.layers.ZeroPadding3D):
+                pass
+            elif isinstance(layer, keras.layers.noise.GaussianDropout) or isinstance(layer, keras.layers.noise.GaussianNoise):
+                pass
+            elif isinstance(layer, keras.layers.Dropout):
+                pass
+            else:
+                outputs = self.get_layout_output_tensors(layer)
+                if len(outputs) > 0:
+                    fn = K.function(inputs, outputs)
+                    Y = fn(input_data_x_sample)[0]
+                    Y = np.squeeze(Y)
 
-                fn = K.function(inputs, self.get_layout_output_tensors(layer))
-                Y = fn(input_data_x_sample)[0]
-                Y = np.squeeze(Y)
+                    if Y.size == 1:
+                        Y = np.array([Y])
 
-                if Y.size == 1:
-                    Y = np.array([Y])
+                    node = self.job_model.get_model_node(layer.name)
 
-                node = self.job_model.get_model_node(layer.name)
-                if node and node['activationFunction'] == 'softmax':
-                    image = self.make_image_from_dense_softmax(Y)
-                else:
-                    image = self.make_image_from_dense(Y)
+                    image = None
+                    if len(Y.shape) > 1:
+                        image = PIL.Image.fromarray(get_layer_vis_square(Y))
+                    elif len(Y.shape) == 1:
+                        if node and node['activationFunction'] == 'softmax':
+                            image = self.make_image_from_dense_softmax(Y)
+                        else:
+                            image = self.make_image_from_dense(Y)
 
-                images.append(JobImage(layer.name, image))
+                    if image:
+                        images.append(JobImage(layer.name, image))
 
         return images
 
     def get_layout_output_tensors(self, layer):
         outputs = []
 
-        for idx, node in enumerate(layer.inbound_nodes):
-            outputs.append(layer.get_output_at(idx))
+        if hasattr(layer, 'inbound_nodes'):
+            for idx, node in enumerate(layer.inbound_nodes):
+                outputs.append(layer.get_output_at(idx))
 
         return outputs
 
     def build_confusion_matrix(self):
         confusion_matrix = {}
 
-        model_has_validation_data = hasattr(
-            self.model, 'validation_data') and self.model.validation_data
+        model_has_validation_data = hasattr(self.model, 'validation_data') and self.model.validation_data
+
         if not model_has_validation_data and not self.trainer.data_validation:
             return confusion_matrix
 
@@ -417,6 +525,10 @@ class KerasLogger(Callback):
     def make_image(self, data):
         from keras.preprocessing.image import array_to_img
         try:
+            if len(data.shape) == 2:
+                # grayscale image, just add once channel
+                data = data.reshape((data.shape[0], data.shape[1], 1))
+
             image = array_to_img(data)
         except:
             return None
