@@ -19,6 +19,7 @@ class Trainer():
         self.samples_per_epoch = 1
         # validation sample count per epoch for generator. same name as in keras fit_generator
         self.nb_val_samples = 1
+        self.nb_val_steps = None
 
         self.data_validation = None
         self.data_train = None
@@ -86,5 +87,8 @@ class Trainer():
 
         self.lock.release()
 
+    def set_job_system_info(self, key, value):
+        self.job_backend.set_system_info(key, value)
+
     def set_job_info(self, key, value):
-        self.job_backend.job_set_info_key(key, value)
+        self.job_backend.set_info(key, value)
