@@ -16,6 +16,13 @@ from io import BytesIO
 import six
 import base64
 import PIL.Image
+import sys
+import msgpack
+
+from aetros.const import JOB_STATUS
+from aetros.logger import GeneralLogger
+from aetros.utils import git
+
 try:
     from cStringIO import StringIO
 except ImportError:
@@ -26,8 +33,7 @@ from aetros.MonitorThread import MonitoringThread
 
 def on_shutdown():
     for job in on_shutdown.started_jobs:
-        if job.running:
-            job.done()
+        job.on_shutdown()
 
 
 on_shutdown.started_jobs = []
@@ -49,6 +55,8 @@ def invalid_json_values(obj):
         return obj.tolist()
     if isinstance(obj, bytes):
         return obj.decode('cp437')
+    if isinstance(obj, map):
+        return list(obj)
     raise TypeError('Invalid data type passed to json encoder: ' + type(obj).__name__)
 
 
@@ -83,21 +91,24 @@ class EventListener:
                 callback(parameter)
 
 
-class Client:
-    def __init__(self, api_host, api_token, event_listener, api_port):
+class BackendClient:
+    def __init__(self, api_host, api_port, event_listener):
         """
 
         :type api_host: string
-        :type api_token: string
+        :type api_key: string
         :type event_listener: EventListener
         :type job_id: integer
         """
-        self.api_token = api_token
-        self.api_host = api_host
-        self.api_port = api_port
+        self.api_host = api_host or os.getenv('API_HOST') or 'trainer.aetros.com'
+        self.api_port = int(api_port or os.getenv('API_PORT') or 8051)
+
         self.event_listener = event_listener
         self.message_id = 0
+
+        self.api_key = None
         self.job_id = None
+
         self.thread_instance = None
         self.s = None
 
@@ -106,13 +117,12 @@ class Client:
         self.queue = []
 
         self.active = False
+        self.external_stopped = False
         self.registered = False
         self.connected = False
-        self.read_buffer = ''
-
-    def start(self, job_id):
-        self.job_id = job_id
+        self.read_unpacker = msgpack.Unpacker(encoding='utf8')
 
+    def start(self):
         self.active = True
 
         if not self.thread_instance:
@@ -120,6 +130,9 @@ class Client:
             self.thread_instance.daemon = True
             self.thread_instance.start()
 
+    def on_connect(self):
+        pass
+
     def connect(self):
         locked = False
 
@@ -132,17 +145,7 @@ class Client:
             self.lock.release()
             locked = False
 
-            self.send_message({'register_job_worker': self.api_token, 'job_id': self.job_id})
-            messages = self.read_full_message(self.s)
-
-            if "JOB_REGISTERED" in messages:
-                self.registered = True
-                print("Connected to %s " % (self.api_host,))
-                self.handle_messages(messages)
-                return True
-
-            print("Registration of job %s failed." % (self.job_id,))
-            return False
+            return self.on_connect()
         except socket.error as error:
             if locked:
                 self.lock.release()
@@ -166,6 +169,8 @@ class Client:
 
         if hasattr(error, 'errno') and hasattr(error, 'message'):
             print("Connection error: %d: %s" % (error.errno, error.message,))
+        elif error:
+            print("Connection error: " + str(error))
         else:
             print("Connection error")
 
@@ -184,37 +189,37 @@ class Client:
 
         while True:
             if self.connected and self.registered:
-                try:
-                    if last_ping < time.time() - 1:
-                        # ping every second
-                        last_ping = time.time()
-                        self.send_message("PING")
+                if last_ping < time.time() - 1:
+                    # ping every second
+                    last_ping = time.time()
+                    self.send_message("PING")
 
-                    if self.connected:
-                        # send pending messages
-                        max_messages = 1
-                        for message in self.queue:
-                            if not message['sending'] and not message['sent']:
-                                max_messages += 1
-                                self.send_message(message)
-                                if max_messages > 10:
-                                    # not too much at once, so we have time to listen for incoming messages
+                # see if we can read something
+                self.lock.acquire()
+                readable, writable, exceptional = select.select([self.s], [self.s], [self.s])
+                self.lock.release()
+
+                if exceptional:
+                    self.connection_error()
+
+                if readable:
+                    messages = self.read(self.s)
+                    if messages is not None:
+                        self.handle_messages(messages)
+
+                if writable:
+                    # send pending messages
+                    self.lock.acquire()
+                    sent_size = 0
+                    for message in self.queue:
+                        if not message['sending'] and not message['sent']:
+                            size = self.send_message(message)
+                            if size is not False:
+                                sent_size += size
+                                if sent_size > 1024 * 1024:
+                                    # not too much at once (max 1MB), so we have time to listen for incoming messages
                                     break
-
-                        # see if we can read something
-                        self.lock.acquire()
-                        readable, writable, exceptional = select.select([self.s], [self.s], [])
-                        self.lock.release()
-                        if exceptional:
-                            self.connection_error()
-
-                        for s in readable:
-                            messages = self.read(s)
-                            if messages:
-                                self.handle_messages(messages)
-
-                except Exception as error:
-                    self.connection_error(error)
+                    self.lock.release()
 
             elif not self.connected and self.active:
                 if not self.connect():
@@ -223,13 +228,17 @@ class Client:
             time.sleep(0.1)
 
     def end(self):
-        #send all missing messages
+        # send all missing messages
+
+        i = 0
         while True:
             if len(self.queue) == 0:
                 break
 
+            i += 1
             time.sleep(0.1)
-
+            if i % 50 == 0:
+                print("[AETROS] we still sync job data. %d messages left." % (len(self.queue),))
 
     def close(self):
         self.active = False
@@ -254,22 +263,20 @@ class Client:
         if isinstance(message, dict):
             message['sending'] = True
 
-        msg = json.dumps(message, default=invalid_json_values)
+        import msgpack
+        msg = msgpack.packb(message, default=invalid_json_values)
 
         try:
-            self.lock.acquire()
-            self.s.sendall(msg + "\n")
-            self.lock.release()
+            self.s.sendall(msg)
 
-            return True
+            return len(msg)
         except:
-            self.lock.release()
             self.connection_error()
             return False
 
     def handle_messages(self, messages):
         for message in messages:
-            if 'handled' in message:
+            if 'handled' in message and message['handled'] is True:
                 for qm in self.queue:
                     if qm['id'] == message['id']:
 
@@ -281,91 +288,128 @@ class Client:
 
                         break
 
-            if 'stop' in message:
+            if not self.external_stopped and 'stop' in message:
+                self.external_stopped = True
                 self.event_listener.fire('stop')
 
-    def read_full_message(self, s):
+    def wait_for_at_least_one_message(self, s):
         """
-        Reads until we receive a message termination (\n)
+        Reads until we receive at least one message we can unpack. Return all found messages.
         """
-        message = ''
+
+        unpacker = msgpack.Unpacker(encoding='utf8')
 
         while True:
             chunk = ''
             try:
                 self.lock.acquire()
-                chunk = s.recv(2048)
+                chunk = s.recv(1024)
             finally:
                 self.lock.release()
 
             if chunk == '':
+                # happens only when connection broke. If nothing is to be received, it hangs instead.
                 self.connection_error()
                 return False
 
-            message += chunk
+            unpacker.feed(chunk)
 
-            message, parsed = parse_message(message)
-            if parsed:
-                return parsed
+            messages = [m for m in unpacker]
+            if messages:
+                return messages
 
     def read(self, s):
         """
-        Reads per call current buffer from network stack. If a full message has been collected (\n retrieved)
-        the message will be parsed and returned. If no message has yet been completley transmitted it returns []
-
-        :return: list
+        Reads from the socket and tries to unpack the message. If successful (because msgpack was able to unpack)
+        then we return that message. Else None. Keep calling .read() when new data is available so we try it 
+        again.
         """
 
-        chunk = ''
-
+        self.lock.acquire()
         try:
-            self.lock.acquire()
-            chunk = s.recv(2048)
+            chunk = s.recv(1024 * 1024)
         finally:
             self.lock.release()
 
         if chunk == '':
+            # socket connection broken
             self.connection_error()
-            return False
+            return None
 
-        self.read_buffer += chunk
+        # self.read_buffer.seek(0, 2) #make sure we write at the end
+        self.read_unpacker.feed(chunk)
 
-        self.read_buffer, parsed = parse_message(self.read_buffer)
-        return parsed
+        # self.read_buffer.seek(0)
+        messages = [m for m in self.read_unpacker]
+
+        return messages if messages else None
 
 
-def start_job(name, api_token=None):
+class JobClient(BackendClient):
+    def configure(self, job_id, api_key):
+        self.job_id = job_id
+        self.api_key = api_key
+
+    def on_connect(self):
+        self.send_message({'register_job_worker': self.api_key, 'job_id': self.job_id})
+        messages = self.wait_for_at_least_one_message(self.s)
+
+        message = messages.pop(0)
+        if isinstance(message, dict) and 'a' in message:
+            if "JOB_ABORTED" in message['a']:
+                print("Job aborted meanwhile. Exiting")
+                self.event_listener.fire('stop')
+                self.active = False
+                return False
+
+            if "JOB_REGISTRATION_FAILED" in message['a']:
+                self.event_listener.fire('registration_failed', {'reason': message['reason']})
+                return False
+
+            if "JOB_REGISTERED" in message['a']:
+                self.registered = True
+                print("Connected to %s " % (self.api_host,))
+                self.event_listener.fire('registration')
+                self.handle_messages(messages)
+                return True
+
+        print("Registration of job %s failed." % (self.job_id,))
+        return False
+
+
+def start_job(name, api_key=None):
     """
     Tries to load the job defined in the AETROS_JOB_ID environment variable.
     If not defined, it creates a new job.
     Starts the job as well.
 
-    :param name: string
-    :param api_token: string
+    :param name: string: model name
+    :param api_key: string
     :return: JobBackend
     """
 
-    job_id = os.getenv('AETROS_JOB_ID')
-    if job_id:
-        job = JobBackend(api_token=api_token)
-        job.load(job_id)
+    id = os.getenv('AETROS_JOB_ID')
+    if id:
+        job = JobBackend(api_key=api_key)
+        job.load(id)
     else:
-        job = create_job(name, api_token)
+        job = create_job(name, api_key)
 
     job.start()
+    job.setup_std_output_logging()
 
     return job
 
 
-def create_job(name, api_token=None):
+def create_job(name, api_key=None):
     """
     Creates a new job.
 
-    :param name: string
-    :param api_token: string
+    :param name: string : model name
+    :param api_key: string
     :return: JobBackend
     """
-    job = JobBackend(api_token=api_token)
+    job = JobBackend(api_key=api_key)
     job.create(name)
     job.load()
 
@@ -376,6 +420,7 @@ class JobLossChannel:
     """
     :type job_backend : JobBackend
     """
+
     def __init__(self, job_backend, name, xaxis=None, yaxis=None, layout=None):
         self.name = name
         self.job_backend = job_backend
@@ -392,7 +437,6 @@ class JobLossChannel:
         self.job_backend.job_add_status('channel', message)
 
     def send(self, x, training_loss, validation_loss):
-
         message = {
             'name': self.name,
             'x': x,
@@ -400,6 +444,7 @@ class JobLossChannel:
         }
         self.job_backend.job_add_status('channel-value', message)
 
+
 class JobImage:
     def __init__(self, id, image, title=None):
         self.id = id
@@ -418,25 +463,47 @@ class JobChannel:
     """
     :type job_backend: JobBackend
     """
-    def __init__(self, job_backend, name, traces=None, main_graph=False,
+
+    def __init__(self, job_backend, name, traces=None,
+                 main=False, kpi=False, max_optimization=True,
                  type=None, xaxis=None, yaxis=None, layout=None):
+        """
+        :param job_backend: JobBakend
+        :param name: str
+        :param traces: None|list : per default create a trace based on "name".
+        :param main: bool : whether this channel is visible in the job list as column for better comparison.
+
+        :param kpi: bool : whether this channel is the KPI (key performance indicator).
+                           Used for hyperparameter optimization. Only one channel can be a kpi. Only first trace used.
+
+        :param max_optimization: bool : whether the optimization maximizes or minmizes the kpi . Use max_optimization=False to
+                                        tell the optimization algorithm that his channel minimizes a kpi, for instance the loss of a model.
+
+        :param type: str : One of JobChannel.NUMBER, JobChannel.TEXT, JobChannel.IMAGE
+        :param xaxis: dict
+        :param yaxis: dict
+        :param layout: dict
+        """
         self.name = name
         self.job_backend = job_backend
 
         if not (isinstance(traces, list) or traces is None):
-            raise Exception('traces can only be None or a list of dicts: [{name: "name", option1: ...}, ...]')
+            raise Exception(
+                'traces can only be None or a list of dicts: [{name: "name", option1: ...}, {name: "name2"}, ...]')
 
         if not traces:
             traces = [{'name': name}]
 
         if isinstance(traces, list) and isinstance(traces[0], six.string_types):
-            traces = map(lambda x: {'name': x}, traces)
+            traces = list(map(lambda x: {'name': x}, traces))
 
         message = {
             'name': name,
             'traces': traces,
             'type': type or JobChannel.NUMBER,
-            'main': main_graph,
+            'main': main,
+            'kpi': kpi,
+            'maxOptimization': max_optimization,
             'xaxis': xaxis,
             'yaxis': yaxis,
             'layout': layout,
@@ -449,8 +516,9 @@ class JobChannel:
             y = [y]
 
         if len(y) != len(self.traces):
-            raise Exception('You tried to set more y values (%d items) then traces available in channel %s (%d traces).' % (
-                len(y), self.name, len(self.traces)))
+            raise Exception(
+                'You tried to set more y values (%d items) then traces available in channel %s (%d traces).' % (
+                    len(y), self.name, len(self.traces)))
 
         message = {
             'name': self.name,
@@ -463,33 +531,89 @@ class JobChannel:
 class JobBackend:
     """
     :type event_listener: EventListener
-    :type api_token: string
+    :type api_key: string
     :type job_id: int
     :type client: Client
     :type job: dict
     """
 
-    def __init__(self, job_id=None, api_token=None):
+    def __init__(self, job_id=None, api_key=None):
         self.event_listener = EventListener()
-        self.api_token = api_token if api_token else os.getenv('API_KEY')
+        self.api_key = api_key if api_key else os.getenv('API_KEY')
+
+        if job_id and '/' in job_id:
+            raise Exception('job_id needs to be a job id, not a model name.')
+
+        on_shutdown.started_jobs.append(self)
 
         self.job_id = job_id
         self.client = None
         self.job = None
+
+        self.last_batch_time = time.time()
+        self.start_time = time.time()
+        self.current_epoch = 0
+        self.current_batch = 0
+        self.total_epochs = 0
+        self.made_batches = 0
+        self.batches_per_second = 0
+
+        # done means: done, abort or crash method has been called.
+        self.ended = False
+
+        # running means: the syncer client is running.
         self.running = False
         self.monitoring_thread = None
+        self.general_logger_stdout = GeneralLogger(job_backend=self)
+        self.general_logger_error = GeneralLogger(job_backend=self, error=True)
 
         self.host = os.getenv('API_HOST')
         self.port = int(os.getenv('API_PORT') or 8051)
         if not self.host or self.host == 'false':
-            self.host = 'aetros.com'
+            self.host = 'trainer.aetros.com'
 
         self.last_progress_call = None
         self.job_ids = []
         self.in_request = False
         self.stop_requested = False
         self.event_listener.on('stop', self.external_stop)
-        self.client = Client(self.host, self.api_token, self.event_listener, self.port)
+        self.event_listener.on('aborted', self.external_aborted)
+        self.event_listener.on('registration', self.on_registration)
+        self.event_listener.on('registration_failed', self.on_registration_failed)
+
+        self.client = JobClient(self.host, self.port, self.event_listener)
+
+    def on_registration_failed(self, params):
+        print("Connecting to AETROS failed. Reasons: %s. Wrong API_KEY='%s'?" % (params['reason'], self.api_key,))
+        print("All monitoring disabled for AETROS. Job may still run.", file=sys.stderr)
+        self.crash('Client could not connect to AETROS. Job may still run. Reason: %s' % (params['reason'],))
+
+    def on_registration(self, params):
+        print("Job %s#%d (%s) started. Open http://%s/job/%s to monitor the training." %
+              (self.model_id, self.job_index, self.job_id, self.host, self.job_id))
+
+    def keras_model_integration(self, model, insights=False, confusion_matrix=False):
+        import aetros.keras
+        aetros.keras.KerasIntegration(self.model_id, model, job_backend=self, insights=insights,
+                                      confusion_matrix=confusion_matrix)
+
+    def external_aborted(self, params):
+        self.ended = True
+        self.running = False
+
+        if self.monitoring_thread:
+            self.monitoring_thread.stop()
+
+        self.client.close()
+
+        os.kill(os.getpid(), signal.SIGINT)
+
+    def setup_std_output_logging(self):
+        """
+        Overwrites sys.stdout and sys.stderr so we can send it additionally to Aetros.
+        """
+        sys.stdout = self.general_logger_stdout
+        sys.stderr = self.general_logger_error
 
     def external_stop(self, params):
         print("Job stopped through AETROS Trainer.")
@@ -497,19 +621,70 @@ class JobBackend:
         self.stop_requested = True
         os.kill(os.getpid(), signal.SIGINT)
 
+    def batch(self, batch, total, size=None):
+        time_diff = time.time() - self.last_batch_time
+        self.made_batches += 1
+
+        if time_diff > 1 or batch == total:  # only each second second or last batch
+            self.set_system_info('currentBatch', batch)
+            self.set_system_info('batchSize', size)
+            self.set_system_info('nb_batches', total)
+
+            self.batches_per_second = self.made_batches / time_diff
+            self.made_batches = 0
+            self.last_batch_time = time.time()
+
+            if size:
+                self.set_system_info('samplesPerSecond', self.batches_per_second * size)
+
+            epochs_per_second = self.batches_per_second / total  # all batches
+            self.set_system_info('epochsPerSecond', epochs_per_second)
+
+            elapsed = time.time() - self.start_time
+            self.set_system_info('elapsed', elapsed)
+
+            if self.total_epochs:
+                eta = 0
+                if batch < total:
+                    # time to end this epoch
+                    eta += (total - batch) / self.batches_per_second
+
+                # time until all epochs are done
+                if self.total_epochs - (self.current_epoch - 1) > 0:
+                    eta += (self.total_epochs - (self.current_epoch - 1)) / epochs_per_second
+
+                self.set_system_info('eta', eta)
+
+        self.current_batch = batch
+
     def progress(self, epoch, total):
-        if self.last_progress_call:
+        self.current_epoch = epoch
+        self.total_epochs = total
+        epoch_limit = False
+
+        if 'maxEpochs' in self.job['config']['settings'] and self.job['config']['settings']['maxEpochs'] > 0:
+            epoch_limit = True
+            self.total_epochs = self.job['config']['settings']['maxEpochs']
+
+        if epoch is not 0 and self.last_progress_call:
             # how long took it since the last call?
             time_per_epoch = time.time() - self.last_progress_call
-            eta = time_per_epoch * (total-epoch)
-            self.set_info('eta', eta)
+            eta = time_per_epoch * (self.total_epochs - epoch)
+            self.set_system_info('eta', eta)
             if time_per_epoch > 0:
-                self.set_info('epochsPerSecond', 1 / time_per_epoch)
+                self.set_system_info('epochsPerSecond', 1 / time_per_epoch)
 
-        self.set_info('epoch', epoch)
-        self.set_info('epochs', total)
+        self.set_system_info('epoch', epoch)
+        self.set_system_info('epochs', self.total_epochs)
         self.last_progress_call = time.time()
 
+        if epoch_limit and self.total_epochs > 0:
+            if epoch >= self.total_epochs:
+                print("\nMaxEpochs of %d/%d reached" % (epoch, self.total_epochs))
+                self.done()
+                os.kill(os.getpid(), signal.SIGINT)
+                return
+
     def create_loss_channel(self, name, xaxis=None, yaxis=None, layout=None):
         """
         :param name: string
@@ -518,77 +693,121 @@ class JobBackend:
 
         return JobLossChannel(self, name, xaxis, yaxis, layout)
 
-    def create_channel(self, name, traces=None, main_graph=False,
+    def create_channel(self, name, traces=None,
+                       main=False, kpi=False, max_optimization=True,
                        type=JobChannel.NUMBER,
                        xaxis=None, yaxis=None, layout=None):
         """
-        :param name: string
-        :param traces: list
-        :param main_graph: bool
-        :param type: string JobChannel.NUMBER, JobChannel.TEXT, JobChannel.IMAGE
-        :return: JobChannel
+        :param name: str
+        :param traces: None|list : per default create a trace based on "name".
+        :param main: bool : whether this channel is visible in the job list as column for better comparison.
+
+        :param kpi: bool : whether this channel is the KPI (key performance indicator).
+                           Used for hyperparameter optimization. Only one channel can be a kpi. Only first trace used.
+
+        :param max_optimization: bool : whether the optimization maximizes or minmizes the kpi . Use max_optimization=False to
+                                        tell the optimization algorithm that his channel minimizes a kpi, for instance the loss of a model.
+
+        :param type: str : One of JobChannel.NUMBER, JobChannel.TEXT, JobChannel.IMAGE
+        :param xaxis: dict
+        :param yaxis: dict
+        :param layout: dict
         """
-        return JobChannel(self, name, traces, main_graph, type, xaxis, yaxis, layout)
+        return JobChannel(self, name, traces, main, kpi, max_optimization, type, xaxis, yaxis, layout)
 
     def start(self):
+        if self.running:
+            raise Exception('Job already running.')
+
         if not self.job_id:
             raise Exception('No job id found. Use create() first.')
 
         self.running = True
-        on_shutdown.started_jobs.append(self)
+        self.ended = False
         self.collect_system_information()
+        self.collect_environment()
         self.start_monitoring()
 
-        self.client.start(self.job_id)
+        self.client.configure(self.job_id, self.api_key)
+        self.client.start()
+        self.detect_git_version()
 
-        print("Job %s#%d (%s) started. Open http://%s/trainer/app#/training=%s to monitor the training." %
-              (self.model_id, self.job_index, self.job_id, self.host, self.job_id))
+    def detect_git_version(self):
+        try:
+            commit_sha = git.get_current_commit_hash()
+            if commit_sha:
+                self.set_system_info('gitVersion', commit_sha)
+
+            current_branch = git.get_current_branch()
+            if current_branch:
+                self.set_system_info('gitBranch', current_branch)
+        except:
+            pass
 
     def start_monitoring(self):
-        self.monitoring_thread = MonitoringThread(self)
-        self.monitoring_thread.daemon = True
-        self.monitoring_thread.start()
+        if not self.monitoring_thread:
+            self.monitoring_thread = MonitoringThread(self)
+            self.monitoring_thread.daemon = True
+            self.monitoring_thread.start()
+
+    def on_shutdown(self):
+        if not self.ended and hasattr(sys, 'last_value'):
+            # sys.last_value contains a exception, when there was an uncaught one
+            if isinstance(sys.last_value, KeyboardInterrupt):
+                self.abort()
+            else:
+                self.crash(type(sys.last_value).__name__ + ': ' + str(sys.last_value))
+
+        elif self.running:
+            self.done()
 
     def done(self):
         if not self.running:
             return
 
+        if self.total_epochs:
+            self.progress(self.total_epochs, self.total_epochs)
+
         self.post('job/stopped', json={'id': self.job_id})
 
-        self.client.end()
-        self.running = False
+        self.stop(True)
 
+    def stop(self, wait_for_client=False):
         if self.monitoring_thread:
             self.monitoring_thread.stop()
 
+        self.ended = True
+        self.running = False
+
+        if wait_for_client:
+            self.client.end()
+        else:
+            self.client.close()
+
     def abort(self):
         if not self.running:
             return
 
         self.post('job/aborted', json={'id': self.job_id})
 
-        self.client.close()
-        self.running = False
-        if self.monitoring_thread:
-            self.monitoring_thread.stop()
+        self.stop()
 
-    def crash(self, e=None):
-        self.post('job/crashed', json={'id': self.job_id, 'error': e.message if e else None})
+    def crash(self, error=None):
+        data = {'id': self.job_id, 'error': str(error) if error else None,
+                'stderr': self.general_logger_error.last_messages}
+        self.post('job/crashed', json=data)
 
-        self.client.close()
-        self.running = False
-        if self.monitoring_thread:
-            self.monitoring_thread.stop()
+        self.stop()
 
     def get_url(self, affix):
 
         url = 'http://%s/api/%s' % (self.host, affix)
 
-        if self.api_token:
+        if self.api_key:
             if '?' in url:
-                url += '&token=' + self.api_token
+                url += '&token=' + self.api_key
             else:
-                url += '?token=' + self.api_token
+                url += '?token=' + self.api_key
 
         return url
 
@@ -708,8 +927,14 @@ class JobBackend:
 
         return requests.put(self.get_url(url), data=data, **kwargs)
 
-    def create(self, name, server_id='local', dataset_id=None, insights=False):
-        response = self.put('job', {'modelId': name, 'serverId': server_id, 'insights': insights, 'datasetId': dataset_id})
+    def create(self, name, server_id='local', hyperparameter=None, dataset_id=None, insights=False):
+        response = self.put('job', json={
+            'modelId': name,
+            'serverId': server_id,
+            'hyperParameters': hyperparameter,
+            'insights': insights,
+            'datasetId': dataset_id
+        })
 
         if response.status_code != 200:
             raise Exception("Could not create job: %s" % (response.content,))
@@ -718,14 +943,37 @@ class JobBackend:
 
         return self.job_id
 
-    def ensure_model(self, name, model_json, settings=None, type='custom', layers=None, graph=None):
+    def ensure_job(self, id, hyperparameter=None, dataset_id=None, server_id='local', insights=False,
+                   insights_sample_path=None, api_key=None):
+        """
+        Makes sure that either
+        1. id is a model name, a job is created and loaded, or
+        2. id is a job id and job is loaded.
+
+        :param id: integer|str: model name or job id.
+        """
+
+        if '/' in id:
+            # it's a model name
+            id = self.create(id, server_id=server_id, hyperparameter=hyperparameter,
+                             dataset_id=dataset_id, insights=insights)
+            self.load(id)
+
+            print(
+                "Job %s#%d (%s) created and started. Open http://%s/job/%s to monitor the training." %
+                (self.model_id, self.job_index, self.job_id, self.host, id))
+        else:
+            self.load(id)
+            if self.job['progressStatus'] > JOB_STATUS.PROGRESS_STATUS_QUEUED:
+                self.restart(id)
+                self.load(id)
+
+    def ensure_model(self, name, settings=None, type='custom', layers=None):
         response = self.put('model/ensure', {
             'id': name,
             'type': type,
-            'model': model_json,
             'settings': json.dumps(settings, allow_nan=False, default=invalid_json_values) if settings else None,
             'layers': json.dumps(layers, allow_nan=False, default=invalid_json_values),
-            'graph': json.dumps(graph, allow_nan=False, default=invalid_json_values),
         })
 
         if response.status_code != 200:
@@ -760,10 +1008,35 @@ class JobBackend:
         if 'hyperParameters' not in self.job['config'] or not self.job['config']['hyperParameters']:
             raise Exception('This job does not have any hyper-parameters')
 
-        if name not in self.job['config']['hyperParameters']:
-            raise Exception('This job does not have the hype parameter %s' % (name,))
+        params = self.job['config']['hyperParameters']
+        if name in params:
+            return params[name]
 
-        return self.job['config']['hyperParameters'][name]
+        if '.' in name:
+            path = name.split('.')
+            current = params
+            for item in path:
+                current = current[item]
+
+            return current
+
+        raise Exception('This job does not have the hype parameter %s' % (name,))
+
+    def restart(self, id=None):
+        """
+        Restarts job.
+        :param id: int
+        """
+        if id:
+            self.job_id = id
+
+        if not self.job_id:
+            raise Exception('No job id given.')
+
+        response = self.post('job/restart', {'id': self.job_id})
+
+        if response.status_code != 200:
+            raise Exception("Could not restart job: %s" % (response.content,))
 
     def load(self, id=None):
         """
@@ -804,7 +1077,10 @@ class JobBackend:
     def sync_weights(self):
         self.job_add_status('status', 'SYNC WEIGHTS')
         print("Sync weights ...")
-        self.upload_weights('best.hdf5', self.get_job_model().get_weights_filepath_best(), with_status=True)
+        try:
+            self.upload_weights('best.hdf5', self.get_job_model().get_weights_filepath_best(), with_status=True)
+        except:
+            pass
         print("Weights synced.")
 
     def load_light_job(self, id=None):
@@ -850,15 +1126,43 @@ class JobBackend:
             'data': item
         })
 
-    def set_info(self, name, value):
-        self.job_set_info_key(name, value)
+    def set_info(self, key, value):
+        self.client.send({
+            'type': 'job-custom-info',
+            'data': {'key': key, 'value': value}
+        })
 
-    def job_set_info_key(self, key, value):
+    def set_graph(self, graph):
+        self.client.send({
+            'type': 'job-graph',
+            'data': {'graph': graph}
+        })
+
+    def set_system_info(self, key, value):
         self.client.send({
             'type': 'job-info',
             'data': {'key': key, 'value': value}
         })
 
+    def upload_file(self, file_path, name=None, mime_type=None):
+        if os.path.getsize(file_path) > 10 * 1024 * 1024 * 1024:
+            raise Exception('Can not upload file bigger than 10MB')
+
+        with open(file_path, 'rb') as f:
+            contents = f.read()
+
+        if mime_type is None:
+            import mimetypes
+            mime_type, encoding = mimetypes.guess_type(file_path)
+
+        path = os.path.relpath(file_path)
+
+        self.client.send({
+            'type': 'job-file',
+            'time': time.time(),
+            'data': {'path': path, 'name': name, 'mime': mime_type, 'content': contents}
+        })
+
     def job_add_insight(self, x, images, confusion_matrix):
         info = {'epoch': x, 'confusionMatrix': confusion_matrix}
 
@@ -870,7 +1174,7 @@ class JobBackend:
             converted_images.append({
                 'id': image.id,
                 'title': image.title,
-                'image': self.to_base64(image.image)
+                'image': self.pil_image_to_jpeg(image.image)
             })
 
         self.client.send({
@@ -879,19 +1183,37 @@ class JobBackend:
             'data': {'info': info, 'images': converted_images}
         })
 
-    def to_base64(self, image):
-        buffer = BytesIO()
-        if (six.PY2):
-            buffer = StringIO()
-        image.save(buffer, format="JPEG", optimize=True, quality=80)
-        return base64.b64encode(buffer.getvalue())
+    def pil_image_to_jpeg(self, image):
+        buffer = six.BytesIO()
 
+        image.save(buffer, format="JPEG", optimize=True, quality=80)
+        return buffer.getvalue()
+
+    def collect_environment(self):
+        import socket
+        import os
+        import pip
+        import platform
+
+        env = {}
+
+        env['python_version'] = platform.python_version()
+        env['python_executable'] = sys.executable
+
+        env['hostname'] = socket.gethostname()
+        env['variables'] = dict(os.environ)
+
+        if 'API_KEY' in env['variables']:
+            del env['variables']['API_KEY']
+
+        env['pip_packages'] = sorted([[i.key, i.version] for i in pip.get_installed_distributions()])
+        self.set_system_info('environment', env)
 
     def collect_system_information(self):
         import psutil
 
         mem = psutil.virtual_memory()
-        self.job_set_info_key('memory_total', mem.total)
+        self.set_system_info('memory_total', mem.total)
 
         on_gpu = False
 
@@ -899,23 +1221,24 @@ class JobBackend:
         if 'theano.sandbox' in sys.modules:
             # at this point, theano is already initialised, so we can use it to monitor the GPU.
             from theano.sandbox import cuda
-            self.job_set_info_key('cuda_available', cuda.cuda_available)
+            self.set_system_info('cuda_available', cuda.cuda_available)
             if cuda.cuda_available:
                 on_gpu = cuda.use.device_number is not None
-                self.job_set_info_key('cuda_device_number', cuda.active_device_number())
-                self.job_set_info_key('cuda_device_name', cuda.active_device_name())
+                self.set_system_info('cuda_device_number', cuda.active_device_number())
+                self.set_system_info('cuda_device_name', cuda.active_device_name())
                 if cuda.cuda_ndarray.cuda_ndarray.mem_info:
                     gpu = cuda.cuda_ndarray.cuda_ndarray.mem_info()
-                    self.job_set_info_key('cuda_device_max_memory', gpu[1])
-                    free = gpu[0] / 1024 / 1024 / 1024
-                    total = gpu[1] / 1024 / 1024 / 1024
+                    self.set_system_info('cuda_device_max_memory', gpu[1])
+                    free = gpu[0] / 1024.0 / 1024.0 / 1024.0
+                    total = gpu[1] / 1024.0 / 1024.0 / 1024.0
                     used = total - free
 
-                    print("%.2fGB GPU memory used of %.2fGB, %s, device id %d" % (used, total, cuda.active_device_name(), cuda.active_device_number()))
+                    print("%.2fGB GPU memory used of %.2fGB, %s, device id %d" % (
+                    used, total, cuda.active_device_name(), cuda.active_device_number()))
 
-        self.job_set_info_key('on_gpu', on_gpu)
+        self.set_system_info('on_gpu', on_gpu)
 
         import cpuinfo
         cpu = cpuinfo.get_cpu_info()
-        self.job_set_info_key('cpu_name', cpu['brand'])
-        self.job_set_info_key('cpu', [cpu['hz_actual_raw'][0], cpu['count']])
+        self.set_system_info('cpu_name', cpu['brand'])
+        self.set_system_info('cpu', [cpu['hz_actual_raw'][0], cpu['count']])
