@@ -14,11 +14,14 @@ class StartCommand:
         parser.add_argument('--insights', action='store_true', help="activates insights. Only for Keras models.")
         parser.add_argument('--insights-sample', help="Path or url to the sample being used to generate the insights. Default is first model sample. Only for Keras models.")
         parser.add_argument('--dataset', help="Dataset id when model has placeholders. Only for Keras models with placeholders as input/output.")
-        parser.add_argument('--secure-key', help="Secure key. Alternatively use API_KEY environment varibale.")
+        parser.add_argument('--api-key', help="Secure key. Alternatively use API_KEY environment varibale.")
         parser.add_argument('--gpu', action='store_true', help="Activates GPU if available. Only for Theano models.")
         parser.add_argument('--device', help="Which device index should be used. Default 0 (which means with --gpu => 'gpu0'). Only for Theano models.")
-        parser.add_argument('--tf', action='store_true', help="Uses TensorFlow instead of Theano. Only for Keras models.")
-        parser.add_argument('--mp', help="Activates multithreading if available with given thread count. Only for Theano models.")
+        parser.add_argument('--tf', action='store_true', help="Force TensorFlow as library. Only for Keras models.")
+        parser.add_argument('--th', action='store_true', help="Force Theano as library. Only for Keras models.")
+        parser.add_argument('--mp', help="Activates multithreading if available with given thread count. Only when Theano is active.")
+        parser.add_argument('--no-hardware-monitoring', action='store_true', help="Deactivates hardware monitoring")
+        parser.add_argument('--param', action='append', help="Sets a hyperparameter, example '--param name=value'. Multiple --param allowed.")
 
         parsed_args = parser.parse_args(args)
 
@@ -39,6 +42,25 @@ class StartCommand:
 
         os.environ['THEANO_FLAGS'] = flags
 
-        os.environ['KERAS_BACKEND'] = 'tensorflow' if parsed_args.tf else 'theano'
+        if parsed_args.tf:
+            os.environ['KERAS_BACKEND'] = 'tensorflow'
 
-        start(parsed_args.name, dataset_id=parsed_args.dataset, insights=parsed_args.insights, insights_sample_path=parsed_args.insights_sample, api_token=parsed_args.secure_key)
+        if parsed_args.th:
+            os.environ['KERAS_BACKEND'] = 'theano'
+
+        hyperparameter = {}
+        if parsed_args.param:
+            for param in parsed_args.param:
+                if '=' not in param:
+                    raise Exception('--param ' + param+' does not contain a =. Please use "--param name=value"')
+
+                name, value = param.split('=')
+                hyperparameter[name] = value
+
+        start(parsed_args.name,
+            hyperparameter=hyperparameter,
+            dataset_id=parsed_args.dataset,
+            insights=parsed_args.insights,
+            insights_sample_path=parsed_args.insights_sample,
+            api_key=parsed_args.api_key,
+        )
